# Copyright (C) 2007-2026 Ahmet Öztürk (aoz_2@yahoo.com)
#
# This file is part of Lifeograph.
#
# Lifeograph is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lifeograph is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

from LoG import *

NAME="Move Paragraphs into Child Entry"
VERSION="0.1"
CATEGORY=SCtg.PARA_MODIFIER

def run(paras):
    # get the parent entry i.e. the current host of the selected paragraphs:
    entry_p = paras[0].get_host()
    #create a child entry:
    entry_c = Diary.get_main().create_entry_child(entry_p)
    # move the pragraphs:
    for p in paras:
        p.move_to_entry(entry_c)
    # remove the empty paragraph added in create_entry_rel:
    entry_c.remove_paragraphs(0)
    entry_p.set_expanded(True)
    refresh_panels()
