# Copyright (C) 2007-2026 Ahmet Öztürk (aoz_2@yahoo.com)
#
# This file is part of Lifeograph.
#
# Lifeograph is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lifeograph is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

from LoG import *
import webbrowser

NAME="Open Location in Google Maps"
VERSION="0.1"
CATEGORY=SCtg.PARA_MODIFIER

def run(paras):
    para = paras[0]
    if para.has_location():
        lat = para.get_latitude()
        lon = para.get_longitude()
        webbrowser.open(f"https://www.google.com/maps?q={lat},{lon}")
    else:
        show_message("Selected paragraph does not have a location!")
