"""
Paragraph Heading Sizes
"""
from __future__ import annotations
__all__: list[str] = ['LARGE', 'MEDIUM', 'NORMAL_TEXT', 'TITLE']
LARGE: int = 261
MEDIUM: int = 260
NORMAL_TEXT: int = 0
TITLE: int = 263
