/***********************************************************************************

    Copyright (C) 2007-2026 Ahmet Öztürk (aoz_2@yahoo.com)

    This file is part of Lifeograph.

    Lifeograph is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Lifeograph is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

***********************************************************************************/


#ifndef LIFEOGRAPH_WIDGET_TEXTVIEW_DIARY_POPUP_HEADER
#define LIFEOGRAPH_WIDGET_TEXTVIEW_DIARY_POPUP_HEADER


#include "widget_textviewedit.hpp"


namespace LoG
{

// TEXTVIEWEDIT ====================================================================================
class TextviewDiaryPopup : public TextviewDiaryEdit
{
    public:
        TextviewDiaryPopup( BaseObjectType* cobject, const Glib::RefPtr< Gtk::Builder >& pb )
        :   TextviewDiaryEdit( cobject, pb, BULLET_COL_WIDTH + 2 ) { }
        // ~TextviewDiaryPopup();

        EditorType                  get_editor_type() const override { return EditorType::POPUP; }

        void                        unset_entry() = delete;

    protected:
        //bool                        on_key_press_event( guint, guint, Gdk::ModifierType ) override;
        //bool                        on_key_release_event(  );

};

} // end of namespace LoG

#endif
