"""
Rich Text Formats
"""
from __future__ import annotations
__all__: list[str] = ['BOLD', 'CODE_COMMENT', 'CODE_KEYWORD', 'CODE_STRING', 'COMMENT', 'DATE', 'FADED', 'HIGHLIGHT', 'ITALIC', 'LINK_EVAL', 'LINK_ID', 'LINK_ONTHEFLY', 'LINK_URI', 'STRIKETHRU', 'SUBSCRIPT', 'SUPERSCRIPT', 'TAG', 'TAG_VALUE', 'TIME', 'UNDERLINE']
BOLD: int = 0
CODE_COMMENT: int = 65618
CODE_KEYWORD: int = 65617
CODE_STRING: int = 65619
COMMENT: int = 65551
DATE: int = 17104915
FADED: int = 5
HIGHLIGHT: int = 2
ITALIC: int = 1
LINK_EVAL: int = 786442
LINK_ID: int = 9175051
LINK_ONTHEFLY: int = 327692
LINK_URI: int = 786441
STRIKETHRU: int = 3
SUBSCRIPT: int = 4194310
SUPERSCRIPT: int = 4194311
TAG: int = 8650760
TAG_VALUE: int = 16842766
TIME: int = 16842768
UNDERLINE: int = 4
