"""
Paragraph List Styles
"""
from __future__ import annotations
__all__: list[str] = ['BULLET', 'CANCELED', 'DONE', 'LETTER', 'LETTER_SMALL', 'NUMBER', 'PLAIN', 'PROGRESSED', 'ROMAN', 'ROMAN_SMALL', 'TODO']
BULLET: int = 592
CANCELED: int = 1600
DONE: int = 1584
LETTER: int = 2672
LETTER_SMALL: int = 2688
NUMBER: int = 2656
PLAIN: int = 0
PROGRESSED: int = 1568
ROMAN: int = 2704
ROMAN_SMALL: int = 2720
TODO: int = 1552
