"""
Element Statuses
"""
from __future__ import annotations
__all__: list[str] = ['AUTO', 'CANCELED', 'DONE', 'PROGRESSED', 'TODO']
AUTO: int = 4096
CANCELED: int = 65536
DONE: int = 32768
PROGRESSED: int = 16384
TODO: int = 8192
