"""
Entry Title Styles
"""
from __future__ import annotations
__all__: list[str] = ['DATE_AND_NAME', 'INHERIT', 'MILESTONE', 'NAME_AND_DESCRIPT', 'NAME_ONLY', 'NUMBER_AND_NAME']
DATE_AND_NAME: int = 3
INHERIT: int = 5
MILESTONE: int = 4
NAME_AND_DESCRIPT: int = 1
NAME_ONLY: int = 0
NUMBER_AND_NAME: int = 2
