"""
Code Languages
"""
from __future__ import annotations
__all__: list[str] = ['BASH', 'CPP', 'FORTRAN', 'GENERIC', 'GO', 'HASKELL', 'HTML', 'JAVA', 'JAVASCRIPT', 'KOTLIN', 'LISP', 'LUA', 'OFF', 'PASCAL', 'PERL', 'PYTHON', 'RUBY', 'RUST', 'SCALA', 'SQL']
BASH: str = 'B'
CPP: str = 'C'
FORTRAN: str = 'F'
GENERIC: str = '*'
GO: str = 'G'
HASKELL: str = 'H'
HTML: str = '<'
JAVA: str = 'J'
JAVASCRIPT: str = 'V'
KOTLIN: str = 'K'
LISP: str = '('
LUA: str = 'U'
OFF: str = '_'
PASCAL: str = 'L'
PERL: str = '$'
PYTHON: str = 'P'
RUBY: str = 'R'
RUST: str = 'T'
SCALA: str = 'S'
SQL: str = 'Q'
