# Copyright (C) 2007-2026 Ahmet Öztürk (aoz_2@yahoo.com)
#
# This file is part of Lifeograph.
#
# Lifeograph is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lifeograph is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

from LoG import *
import webbrowser
import urllib.parse

NAME="Lookup in Wikipedia"
VERSION="0.1"
CATEGORY=SCtg.WORD_MODIFIER

def run(para: Paragraph, pos_bgn, pos_end):
    text = para.get_text()

    # if no selection → expand to word boundaries
    if pos_bgn == pos_end:
        delimiters = " \t\n.,;:!?()[]{}<>\"'`"

        # find start
        start = pos_bgn
        while start > 0 and text[start - 1] not in delimiters:
            start -= 1

        # find end
        end = pos_end
        while end < len(text) and text[end] not in delimiters:
            end += 1

        word = text[start:end]
    else:
        word = text[pos_bgn:pos_end]

    word = word.strip()
    if not word:
        return

    # encode for URL
    encoded = urllib.parse.quote(word)

    # set lookup site url:
    url = f"https://en.wikipedia.org/wiki/{encoded}"

    # alternatives:
    # url = f"https://en.wiktionary.org/wiki/{encoded}"
    # url = f"https://dictionary.cambridge.org/dictionary/english/{encoded}"
    # url = f"https://www.merriam-webster.com/dictionary/{encoded}"
    # url = f"https://www.oxfordlearnersdictionaries.com/definition/english/{encoded}"

    webbrowser.open(url)
