"""
Paragraph Styles
"""
from __future__ import annotations
__all__: list[str] = ['IMAGE_CHART', 'IMAGE_FILE', 'IMAGE_TABLE']
IMAGE_CHART: int = 1179648
IMAGE_FILE: int = 1114112
IMAGE_TABLE: int = 1245184
