# Copyright (C) 2007-2026 Ahmet Öztürk (aoz_2@yahoo.com)
#
# This file is part of Lifeograph.
#
# Lifeograph is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lifeograph is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

from LoG import *
from pathlib import Path
import markdown_parsing as mdp

NAME="Import Markdown"
VERSION="0.2"
CATEGORY=SCtg.IMPORTER

def run(path):
    entry = Diary.get_main().create_entry(VT.ETS.NAME_ONLY, Path(path).stem)

    with open(path, encoding="utf-8") as f:
        raw_lines = f.read()

    mdp.import_markdown(entry, raw_lines)

    refresh_panels()

    return True

