#!/bin/bash

for f in "$@"
do
    tmp=$(mktemp)
    scour -i "$f" -o "$tmp" \
        --enable-viewboxing \
        --remove-metadata \
        --strip-xml-prolog \
        --remove-descriptive-elements \
        --indent=none
    mv "$tmp" "$f"
done
